/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Fri Jan 17 17:59:14 UTC 2003
    copyright            : (C) 2003 by Oleksiy Kohany
    email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <stdlib.h>
#include <string.h>
#include "BioSequence.h"
#include "BioAlignment.h"
#include "PairwiseAligner.h"
#include "IOWrapper.h"

#include <boost/program_options.hpp>
namespace po = boost::program_options;

using namespace std;

/** ======================================================== */
void showVersion(const char* exec)
{
   if (!exec) return;
   if (strrchr(exec, '/')) exec = strrchr(exec, '/') + 1;
   cerr << exec;
#ifdef VERSION

   cerr << "\tversion " << VERSION;
#endif

   cerr << endl;
}

/** ======================================================== */
void usage(char* exec, po::options_description od)
{
   if (!exec) return;
   if (strrchr(exec, '/')) exec = strrchr(exec, '/') + 1;
   cerr << "NAME\n\t" << exec << " - aligns pairs of sequences read from input file" << endl << endl;
   cerr << "SYNOPSYS\n\t" << exec << " infile(-) [outfile] [options]\n";
   cerr << od;
   cerr << "\nAUTHOR\n\tOleksiy Kohany, kohany@girinst.org\n";
#ifdef COPYRIGHT

   cerr << "\nCOPYRIGHT\n\t" << COPYRIGHT << "\n";
#endif
}

/** ======================================================== */
int main(int argc, char *argv[])
{
   string infile;
   string outfile;

   PairwiseAligner aligner;
   int iGI = 0;
   int iGX = 0;
   string mask;

   po::options_description generic("GENERIC OPTIONS");
   po::options_description settings("SETTINGS");
   po::options_description alignment("ALIGNMENT OPTIONS");
   po::options_description backend("BACKEND OPTIONS");
   po::options_description hidden("HIDDEN OPTIONS");

   generic.add_options()
   ("version,v", "print version string.")
   ("help", "produce help message.")
   //("debug", "will show debugging messages on execution process.")
   ;

   settings.add_options()
   ("protein,p", "will assume that sequences are proteins.")
   ("msk", po::value<string>(&mask), "pattern that should not be included in calculating similarity, if found in bottom sequence.")
   ;

   alignment.add_options()
#ifndef _OPEN_SOURCE_
   ("pgsw", "uses Phil Green's implementation of SWAT algorithm. Default matrix - xrepmask, gi=30, gx=6")
#endif
#ifdef HAVE_SSE2
   ("sse2", "uses modified implementation of Michael Farrar's striped SWAT algorithm for SSE2 extension of x86 CPUs. This is the default alignment method. Default matrix - xrepmask/3, gi=-10, gx=-2. Limitations: Maximum score can not accede 65535 so keep alignment matrix's scores low. With the default nucleotide matrix you can align over 21KB.")
#endif
   ("mon", "use Monardo alignment algorithm. It has logarithmic gap penalty though much slower than sse2 version.")
   ("gi", po::value<int>(&iGI), "gap initialization penalty")
   ("gx", po::value<int>(&iGX), "gap extension penalty")
   ("mat", po::value<string>(), "file name for alignment scoring matrix")
   ;

   backend.add_options()
   //("nothreads","do not use multithreading. Useful if for some reason you need to reduce CPU utilization to just 1 processor. By default if -mon or -sse2 option is used the program will attempt to utilize all available CPUs for multiple alignment stage. Phill Green's SWAT implementation is not thread safe so it uses just one CPU thread.")
   ;
   hidden.add_options()
   ("input-file", po::value<string>(&infile), "top sequence file.")
   ("output-file", po::value<string>(&outfile)->default_value("-"), "output file. If '-' - stdout will be taken")
   ;

   po::options_description cmdline_options;
   cmdline_options.add(generic).add(settings).add(alignment).add(backend).add(hidden);

   po::options_description visible_options;
   visible_options.add(generic).add(settings).add(alignment).add(backend);

   po::positional_options_description p;
   p.add("input-file", 1).add("output-file", 1);

   po::variables_map vm;
   po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(p).run(), vm);
   po::notify(vm);

   if (vm.count("version")) {
      showVersion(argv[0]);
      exit(0);
   }
   if (vm.count("input-file") == 0) {
      usage(argv[0], visible_options);
      exit(0);
   }
   if (vm.count("debug")) {
      //DEBUG   = true;
   }
  if (vm.count("pgsw")) {
      aligner.setAlignmentMethod(PGSWAT);
   }
   if (vm.count("mon")) {
      aligner.setAlignmentMethod(MONARDO);
   }
   if (vm.count("sse2")) {
      aligner.setAlignmentMethod(SWSSE2);
   }
   if (vm.count("protein")) {
      BioAlignment::loadMatrix("/usr/local/lib/mat/BLOSUM62");
      aligner.setAlignmentMatrix("/usr/local/lib/mat/BLOSUM62");
   }
   if (vm.count("mat")) {
      BioAlignment::loadMatrix(vm["mat"].as<string> ().c_str());
      aligner.setAlignmentMatrix(vm["mat"].as<string> ().c_str());
   }

   if (iGI != 0) {
      aligner.setGapIni(iGI);
   }
   if (iGX != 0) {
      aligner.setGapExt(iGX);
   }

   IstreamWrapper is(infile.c_str());
   OstreamWrapper os(outfile.c_str());
   //========================================================
   BioAlignment aln;
   if (vm.count("msk")) {
      aln.addMask(vm["msk"].as<string> ().c_str());
   }
   while (aln.getSeq().read(is) > 0 && aln.getCon().read(is) > 0) {
      if (aligner.align(aln) == 0) {
         aln.getCon().assign("");
         aln.getSeq().assign("");
      }
      aln.print(os);
      os.stream() << ";" << endl;
   }
}
