/***************************************************************************
 main.cpp  -  description
 -------------------
 begin                : Fri Jan 17 17:59:14 UTC 2003
 copyright            : (C) 2003 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string.h>
#include <list>
#include <exception>
#include "MultipleAlignment.h"
#include "BioSequence.h"
#include "IOWrapper.h"

#include <boost/program_options.hpp>
namespace po = boost::program_options;

using namespace std;

typedef list<BioSequence> SEQ_LIST;

/** ======================================================== */
void usage(char* exec, po::options_description od) {
   if (!exec)
      return;
   if (strrchr(exec, '/'))
      exec = strrchr(exec, '/') + 1;
   cerr << "NAME\n\t" << exec << " - creates a multiple alignment to a reference sequence.\n\n";
   cerr << "SYNOPSYS\n\t" << exec << " refseq|- seqset|- [outfile] [options]\n\n";
   cerr << "DESCRIPTION\n" << "\tspecifying '-' instead of input file name will make the program\n";
   cerr << "\tto read from standard input. If both inputs are taken from stdin\n";
   cerr << "\tthe first sequence will become the reference.\n";
   cerr << od;
   cerr << "\nAUTHOR\n\tOleksiy Kohany, kohany@girinst.org\n";
#ifdef COPYRIGHT

   cerr << "\nCOPYRIGHT\n\t" << COPYRIGHT << "\n";
#endif
}
/** ======================================================== */
void showVersion(const char* exec) {
   if (!exec)
      return;
   if (strrchr(exec, '/'))
      exec = strrchr(exec, '/') + 1;
   cerr << exec;
#ifdef VERSION

   cerr << "\tversion " << VERSION;
#endif

   cerr << endl;
}

/** ======================================================== */
int main(int argc, char *argv[]) {
   string infile1;
   string infile2;
   string outfile;

   int DEBUG = 0;
   int iGI = 0;
   int iGX = 0;
   bool noref = false;

   po::options_description generic("GENERIC OPTIONS");
   po::options_description settings("SETTINGS");
   po::options_description alignment("ALIGNMENT OPTIONS");
   po::options_description backend("BACKEND OPTIONS");
   po::options_description hidden("HIDDEN OPTIONS");

   generic.add_options()
   ("version,v", "print version string.")
   ("help", "produce help message.")
   //("debug", "will show debugging messages on execution process.")
   ;

   settings.add_options()
   ("protein,p", "will assume that sequences are proteins.")
   ("noref", "will remove reference sequence from the multiple alignment")
   ("nogappad", "flanking regions will be padded with '<', '>' instead of '-' to a complete alignment width.")
   ;

   alignment.add_options()
#ifndef _OPEN_SOURCE_
   ("pgsw", "uses Phil Green's implementation of SWAT algorithm. Default matrix - xrepmask, gi=30, gx=6")
#endif
#ifdef __SSE2__
   ("sse2", "uses modified implementation of Michael Farrar's striped SWAT algorithm for SSE2 extension of x86 CPUs. This is the default alignment method. Default matrix - xrepmask/3, gi=-10, gx=-2. Limitations: Maximum score can not accede 65535 so keep alignment matrix's scores low. With the default nucleotide matrix you can align over 21KB.")
#endif
   ("mon", "use Monardo alignment algorithm. It has logarithmic gap penalty though much slower than sse2 version.")
   ("gi", po::value<int>(&iGI), "gap initialization penalty")
   ("gx", po::value<int>(&iGX), "gap extension penalty")
   ("mat", po::value<string>(), "file name for alignment scoring matrix")
   ;

   backend.add_options()
   ("nothreads", "do not use multithreading. Useful if for some reason you need to reduce CPU utilization to just 1 core.");
   hidden.add_options()
   ("input-file1", po::value<string>(&infile1), "input file1. If '-' - stdin will be taken")
   ("input-file2", po::value<string>(&infile2), "input file2.")
   ("output-file", po::value<string>(&outfile)->default_value("-"), "output file. If '-' - stdout will be taken")
   ;

   po::options_description cmdline_options;
   cmdline_options.add(generic).add(settings).add(alignment).add(backend).add(hidden);

   po::options_description visible_options;
   visible_options.add(generic).add(settings).add(alignment).add(backend);

   po::positional_options_description p;
   p.add("input-file1", 1).add("input-file2", 1).add("output-file", 1);

   po::variables_map vm;
   po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(p).run(), vm);
   po::notify(vm);

   if (vm.count("version")) {
      showVersion(argv[0]);
      exit(0);
   }
   if (vm.count("input-file1") == 0 || vm.count("input-file2") == 0) {
      usage(argv[0], visible_options);
      exit(0);
   }
   if (vm.count("debug")) {
      //DEBUG   = true;
   }
   if (vm.count("pgsw")) {
      MultipleAlignment::setAlignmentMethod( PGSWAT);
   }
   if (vm.count("mon")) {
      MultipleAlignment::setAlignmentMethod( MONARDO);
   }
   if (vm.count("sse2")) {
      MultipleAlignment::setAlignmentMethod( SWSSE2);
   }
   if (vm.count("nothreads")) {
      ThreadedAligner::setMultithreaded(false);
   }
   if (vm.count("noref")) {
      noref = true;
   }
   if (vm.count("protein")) {
      BioAlignment::loadMatrix("/usr/local/lib/mat/BLOSUM62");
      MultipleAlignment::setAlignmentMatrix("/usr/local/lib/mat/BLOSUM62");
   }
   if (vm.count("mat")) {
      BioAlignment::loadMatrix(vm["mat"].as<string> ().c_str());
      MultipleAlignment::setAlignmentMatrix(vm["mat"].as<string> ().c_str());
   }

   if (iGI != 0) {
      MultipleAlignment::setGapIni(iGI);
   }
   if (iGX != 0) {
      MultipleAlignment::setGapExt(iGX);
   }

   IstreamWrapper is1(infile1.c_str());
   IstreamWrapper is2(infile2.c_str());
   OstreamWrapper os(outfile.c_str());
   //========================================================
   SEQ_LIST seqls;
   BioSequence sREF;
   BioSequence sSEQ;
   if (sREF.read(is1) > 0) {
      while (sSEQ.read(is2) >= 0) {
         seqls.push_back(sSEQ);
      }
      MultipleAlignment maln;
      maln.construct(sREF, seqls.begin(), seqls.end());
      if (noref) {
         if (maln.size() > 0) {
            maln.erase(maln.begin());
         }
      }
      if (vm.count("nogappad") == 0) {
         for (MultipleAlignment::iterator vptr = maln.begin(); vptr != maln.end(); vptr++) {
            for (BioSequence::iterator sptr = (*vptr).begin(); sptr != (*vptr).end(); sptr++) {
               if (*sptr == ScoreMatrix::RPADDING || *sptr == ScoreMatrix::LPADDING) {
                  *sptr = ScoreMatrix::GAPCHAR;
               }
            }
         }
      }
      maln.print(os);
   }
}
